/*	$NetBSD$	*/

/*-
 * Copyright (c) 2000 Citrus Project,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
#if defined(LIBC_SCCS) && !defined(lint)
__RCSID("$NetBSD$");
#endif /* LIBC_SCCS and not lint */

#include <sys/types.h>
#include <sys/param.h>
#include <errno.h>
#include <iconv.h>
#include <string.h>
#include <stdlib.h>
#include "setlocale.h"
#ifndef __FreeBSD__
#define _BSD_RUNE_T_    wchar_t
#define _BSD_CT_RUNE_T_ wchar_t
#include "rune.h"
#include "runetype.h"
#else
#include <rune.h>
#include <runetype.h>
#endif
#include "rune_local.h"
#include "iconv_local.h"

#ifndef LC_CTYPE
#define LC_CTYPE	2	/*XXX*/
#endif

iconv_t
iconv_open(tocode, fromcode)
	const char *tocode;
	const char *fromcode;
{
	struct __iconv_local *p;
	char path[PATH_MAX];

	p = malloc(sizeof(*p));
	if (!p) {
		errno = ENOMEM;
		goto fail;
	}

	/* XXX should use special pathname for locale database */
	snprintf(path, sizeof(path), "%s/%s/LC_CTYPE", _PATH_LOCALE,
	    tocode);
	/* LINTED const cast */
	if (_newrunelocale(path) != 0 ||
	    (p->tolocale = _findrunelocale(path)) == NULL) {
		errno = EINVAL;
		free(p);
		goto fail;
	}

	/* XXX ditto */
	snprintf(path, sizeof(path), "%s/%s/LC_CTYPE", _PATH_LOCALE,
	    fromcode);
	/* LINTED const cast */
	if (_newrunelocale(path) != 0 ||
	    (p->fromlocale = _findrunelocale(path)) == NULL) {
#if 0
		_NukeRune(p->tolocale)
#endif
		errno = EINVAL;
		free(p);
		goto fail;
	}

	/*
	 * XXX here, we should do either of the following:
	 * - if p->{to,from}locale->__encoding are equal, do nothing.
	 * - if they are not equal, we need to grab conversion table.
	 *   (like between iso2022.c wchar_t encoding to UCS4)
	 * - if no conversion table was found, bail.
	 */
	if (strcmp(p->tolocale->__encoding, p->fromlocale->__encoding) != 0) {
#if 0
		_NukeRune(p->tolocale)
		_NukeRune(p->fromlocale)
#endif
		errno = EINVAL;
		free(p);
		goto fail;
	}

	_mbsinit_rl(&p->tostate, p->tolocale);
	_mbsinit_rl(&p->fromstate, p->fromlocale);

	return (iconv_t)p;

fail:
	return (iconv_t)-1;
}

size_t
iconv(cd, inbuf, inbytesleft, outbuf, outbytesleft)
	iconv_t cd;
	const char **inbuf;
	size_t *inbytesleft;
	char **outbuf;
	size_t *outbytesleft;
{
	struct __iconv_local *p;
	size_t cnt;
	size_t l;
	size_t wl;
	wchar_t *w = NULL;
	const char *cp;
	const wchar_t *wcp;

	if (cd == (iconv_t)-1) {
		errno = EBADF;
		goto fail;
	}

	p = (struct __iconv_local *)cd;

	if (!inbuf || !*inbuf) {
		if (outbytesleft && *outbytesleft && *outbytesleft > 0) {
			/* should emit initialization sequence into outbuf */
			return 0;
		}

		errno = EINVAL;
		goto fail;
	}

	if (!inbytesleft || !outbytesleft) {
		errno = EINVAL;
		goto fail;
	}
	if (!outbuf || !*outbuf) {
		errno = EINVAL;
		goto fail;
	}

	w = malloc(sizeof(*w) * (*inbytesleft + 1));
	if (!w) {
		errno = ENOMEM;
		goto fail;
	}

	cnt = 0;
	while (*inbytesleft > 0 && *outbytesleft > 0) {
		/* speical case - copy \0 as is */
		if (!**inbuf) {
			**outbuf = **inbuf;
			(*inbuf)++;
			(*outbuf)++;
			(*inbytesleft)--;
			(*outbytesleft)--;
			continue;
		}

		wl = *inbytesleft;
		memset(w, 0, sizeof(*w) * (wl + 1));

		/* convert inbuf into wchar */
		cp = *inbuf;
		l = _mbsrtowcs_rl(w, inbuf, wl, &p->fromstate, p->fromlocale);
		if (l < 0) {
			errno = EILSEQ;
			goto fail;
		} else if (l > wl) {
			errno = ENOMEM;
			goto fail;
		}
		w[l] = (wchar_t)0;	/* ensure termination */
		*inbytesleft -= (*inbuf - cp);

		/* XXX consult wchar_t conversion table here, if we have one. */

		/* convert wchar to outbuf */
		wcp = w;
		l = _wcsrtombs_rl(*outbuf, (const wchar_t **)&w, *outbytesleft,
		    &p->tostate, p->tolocale);
		if (l < 0) {
			errno = EILSEQ;
			goto fail;
		}
		*outbuf += l;
		*outbytesleft -= l;

		cnt += l;	/*???*/
	}

	if (w)
		free(w);
	return cnt;

fail:
	if (w)
		free(w);
	return (size_t)-1;
}

int
iconv_close(cd)
	iconv_t cd;
{
	struct __iconv_local *p;

	if (cd == (iconv_t)-1) {
		errno = EBADF;
		return -1;
	}

	if (cd != NULL) {
		p = (struct __iconv_local *)cd;
#if 0
		_NukeRune(p->tolocale)
		_NukeRune(p->fromlocale);
#endif
		free(p);
	}
	/* just in case */
	memset(&p->tostate, 0, sizeof(p->tostate));
	memset(&p->fromstate, 0, sizeof(p->fromstate));

	return 0;
}
