/*	$NetBSD: setlocale.c,v 1.17 1999/10/15 17:17:07 jdolecek Exp $	*/

/*
 * Copyright (c) 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Paul Borman at Krystal Technologies.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
#if defined(LIBC_SCCS) && !defined(lint)
#if 0
static char sccsid[] = "@(#)setlocale.c	8.1 (Berkeley) 7/4/93";
#else
#ifndef __RCSID
#define __RCSID(a)
#endif
__RCSID("$NetBSD: setlocale.c,v 1.17 1999/10/15 17:17:07 jdolecek Exp $");
#endif
#endif /* LIBC_SCCS and not lint */

#define _CTYPE_PRIVATE

#include "namespace.h"
#include <sys/localedef.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#define __SETLOCALE_SOURCE__
#include <locale.h>
#include <paths.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifndef WITH_RUNE
#include "ctypeio.h"
#else
#include <sys/stat.h>
#include "collate.h"
#include "timelocal.h"
size_t __mb_len_max_runtime = MB_LEN_MAX;
#endif

#define	_BSD_RUNE_T_	int
#define _BSD_CT_RUNE_T_	rune_t
#include "rune.h"
#include "rune_local.h"

/*
 * Category names for getenv()
 */
static const char *const categories[_LC_LAST] = {
    "LC_ALL",
    "LC_COLLATE",
    "LC_CTYPE",
    "LC_MONETARY",
    "LC_NUMERIC",
    "LC_TIME",
    "LC_MESSAGES"
};

/*
 * Current locales for each category
 */
static char current_categories[_LC_LAST][32] = {
    "C",
    "C",
    "C",
    "C",
    "C",
    "C",
    "C"
};

/*
 * The locales we are going to try and load
 */
static char new_categories[_LC_LAST][32];

/*
 * Backup area to back out changes on failure
 */
static char saved_categories[_LC_LAST][32];

static char current_locale_string[_LC_LAST * 33];
#ifndef WITH_RUNE
static char *PathLocale;
/*
 * Data paths for each category.
 */
static char **paths[_LC_LAST] = {
    &PathLocale,			/* LC_ALL */
    &PathLocale,			/* LC_COLLATE */
    &PathLocale,			/* LC_CTYPE */
    &PathLocale,			/* LC_MONETARY */
    &PathLocale,			/* LC_NUMERIC */
    &PathLocale,			/* LC_TIME */
    &PathLocale,			/* LC_MESSAGES */
};
#else
#include "setlocale.h"
#endif

static char	*currentlocale __P((void));
static char	*loadlocale __P((int));
#ifdef WITH_RUNE
static int	stub_load_locale __P((const char *, int));
#endif


char *
__setlocale(category, locale)
	int category;
	const char *locale;
{
	int i, loadlocale_success;
	size_t len;
	char *env, *r;

#ifndef WITH_RUNE
	/*
	 * XXX potential security problem here with set-id programs
	 * being able to read files the user can not normally read.
	 */
	if (!PathLocale) {
		if(getenv("PATH_LOCALE")) {
			PathLocale = strdup(getenv("PATH_LOCALE"));
		} else {
			PathLocale = _PATH_LOCALE;
		}
	}
#else
	if (_PathLocale == NULL) {
		char *p = getenv("PATH_LOCALE");

		if (p != NULL && !issetugid()) {
			if (strlen(p) + 1/*"/"*/ + ENCODING_LEN +
			    1/*"/"*/ + CATEGORY_LEN >= PATH_MAX) {
				errno = EFAULT;
				return(NULL);
			}
			_PathLocale = strdup(p);
			if (_PathLocale == NULL)
				return (NULL);
		} else {
			_PathLocale = _PATH_LOCALE;
		}
	}
#endif

	if (category < 0 || category >= _LC_LAST)
		return (NULL);

	if (!locale)
		return (category ?
		    current_categories[category] : currentlocale());

	/*
	 * Default to the current locale for everything.
	 */
	for (i = 1; i < _LC_LAST; ++i)
		(void)strlcpy(new_categories[i], current_categories[i],
		    sizeof(new_categories[i]));

	/*
	 * Now go fill up new_categories from the locale argument
	 */
	if (!*locale) {
		env = getenv(categories[category]);

		if (!env || !*env)
			env = getenv(categories[0]);

		if (!env || !*env)
			env = getenv("LANG");

		if (!env || !*env || strchr(env, '/'))
			env = "C";

		(void)strlcpy(new_categories[category], env,
		    sizeof(new_categories[category]));
		if (!category) {
			for (i = 1; i < _LC_LAST; ++i) {
				env = getenv(categories[i]);
				if (!env || !*env || strchr(env, '/'))
					env = new_categories[0];
				(void)strlcpy(new_categories[i], env,
				    sizeof(new_categories[i]));
			}
		}
	} else if (category) {
		(void)strlcpy(new_categories[category], locale,
		    sizeof(new_categories[category]));
	} else {
		if ((r = strchr(locale, '/')) == 0) {
			for (i = 1; i < _LC_LAST; ++i) {
				(void)strlcpy(new_categories[i], locale,
				    sizeof(new_categories[i]));
			}
		} else {
			for (i = 1; r[1] == '/'; ++r)
				;
			if (!r[1])
				return (NULL);	/* Hmm, just slashes... */
			do {
				len = r - locale > sizeof(new_categories[i]) - 1
					? sizeof(new_categories[i]) - 1
					: r - locale;
				(void)strncpy(new_categories[i++], locale, len);
				new_categories[i++][len] = 0;
				locale = r;
				while (*locale == '/')
				    ++locale;
				while (*++r && *r != '/');
			} while (*locale);
			while (i < _LC_LAST)
				(void)strlcpy(new_categories[i],
				    new_categories[i - 1],
				    sizeof(new_categories[i]));
		}
	}

	if (category)
		return (loadlocale(category));

	loadlocale_success = 0;
	for (i = 1; i < _LC_LAST; ++i) {
		(void)strlcpy(saved_categories[i], current_categories[i],
		    sizeof(saved_categories[i]));
		if (loadlocale(i) != NULL)
			loadlocale_success = 1;
	}

	/*
	 * If all categories failed, return NULL; we don't need to back
	 * changes off, since none happened.
	 */
	if (!loadlocale_success)
		return NULL;

	return (currentlocale());
}

static char *
currentlocale()
{
	int i;

	(void)strlcpy(current_locale_string, current_categories[1],
	    sizeof(current_locale_string));

	for (i = 2; i < _LC_LAST; ++i)
		if (strcmp(current_categories[1], current_categories[i])) {
			(void)snprintf(current_locale_string,
			    sizeof(current_locale_string), "%s/%s/%s/%s/%s",
			    current_categories[1], current_categories[2],
			    current_categories[3], current_categories[4],
			    current_categories[5]);
			break;
		}
	return (current_locale_string);
}

static char *
loadlocale(category)
	int category;
{
#ifndef WITH_RUNE
	char name[PATH_MAX];
#endif

	if (strcmp(new_categories[category], current_categories[category]) == 0)
		return (current_categories[category]);

	if (!strcmp(new_categories[category], "C") ||
	    !strcmp(new_categories[category], "POSIX")) {

		switch (category) {
		case LC_CTYPE:
			(void)_xpg4_setrunelocale("C");
			(void)__runetable_to_netbsd_ctype("C");
		}

		(void)strlcpy(current_categories[category],
		    new_categories[category],
		    sizeof(current_categories[category]));
		return current_categories[category];
	}

#ifndef WITH_RUNE
	/*
	 * Some day we will actually look at this file.
	 */
	(void)snprintf(name, sizeof(name), "%s/%s/%s",
	    *paths[category],
	    new_categories[category], categories[category]);
#endif

	switch (category) {
	case LC_CTYPE:
		if (_xpg4_setrunelocale(new_categories[category]))
			return NULL;
		if (__runetable_to_netbsd_ctype(new_categories[category])) {
			/* very unfortunate, but need to go to "C" locale */
			(void)_xpg4_setrunelocale("C");
			(void)__runetable_to_netbsd_ctype("C");
			return NULL;
		}
		(void)strlcpy(current_categories[category],
		    new_categories[category],
		    sizeof(current_categories[category]));
		return current_categories[category];

	case LC_COLLATE:
#ifdef WITH_RUNE
		if (__collate_load_tables(new_categories[category]) == 0) {
			(void)strlcpy(current_categories[category],
			    new_categories[category],
			    sizeof(current_categories[category]));
			return new_categories[category];
		}
		(void)__collate_load_tables(current_categories[category]);
		return NULL;
#else
		/* FALLTHROUGH */
#endif
	case LC_TIME:
#ifdef WITH_RUNE
		if (__time_load_locale(new_categories[category]) == 0) {
			(void)strlcpy(current_categories[category],
			    new_categories[category],
			    sizeof(current_categories[category]));
			return new_categories[category];
		}
		(void)__time_load_locale(current_categories[category]);
		return NULL;

#else
		/* FALLTHROUGH */
#endif
	case LC_MESSAGES:
	case LC_MONETARY:
	case LC_NUMERIC:
#ifdef WITH_RUNE
		if (stub_load_locale(new_categories[category], category) == 0) {
			(void)strlcpy(current_categories[category],
			    new_categories[category],
			    sizeof(current_categories[category]));
			return new_categories[category];
		}
		stub_load_locale(current_categories[category], category);
		return NULL;
#else
		return NULL;
#endif
	}

	return NULL;
}

#ifdef WITH_RUNE
/*
 * Temporary code until we implement actual LC_* code.
 */
static int
stub_load_locale(encoding, category)
 const char *encoding;
 int category;
{
	char name[PATH_MAX];
	struct stat st;

	if (!encoding)
		return(1);
	/*
	 * The "C" and "POSIX" locale are always here.
	 */
	if (!strcmp(encoding, "C") || !strcmp(encoding, "POSIX"))
		return(0);
	if (!_PathLocale)	/* XXX: shouldn't we initialize this? */
		return(1);
	/* Range checking not needed, encoding has fixed size */
	strcpy(name, *_LocalePaths[category]);
	strcat(name, "/");
	strcat(name, encoding);
#if 0
	/*
	 * Some day we will actually look at this file.
	 */
#endif
	return (stat(name, &st) != 0 || !S_ISDIR(st.st_mode));
}
#endif
